/*
 * Copyright (C) 2006 TiVo Inc. All Rights Reserved.
 */ 

#include <asm/pgtable-bits.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/fastbufs.h>

#define ENABLE_FASTBUFS 0

#define dprintk(args...)

extern int remap_pages(unsigned long address, phys_t phys_addr,
	                            phys_t size, unsigned long flags);

#define FBUF_MAX_ENTS           256
struct fbuf_zone{
#define FBUF_MAX_ZBMP_BYTES (FBUF_MAX_ENTS/(8*sizeof(unsigned char)))
    unsigned char bitmap[FBUF_MAX_ZBMP_BYTES]; 
    void    *objp;
    int in_use;
    unsigned long  zone_start,zone_end;
};

static struct fbuf_zone *fbuf_zones;

struct fbuf_ent_sz_t {
#define THREAD_EXEC_ENTRY_SZ    (sizeof(struct thread_exec_entry))
#define CNTXT_ENTRY_SZ          (sizeof(struct context_entry)) 
    int sz;
    int count;
};

static struct fbuf_ent_sz_t fbuf_ent_sz[FBUF_MAX_TYPES] = { 
                    {THREAD_EXEC_ENTRY_SZ, FBUF_MAX_ENTS},
                    {CNTXT_ENTRY_SZ,FBUF_MAX_ENTS},
};

int fastbuf_initialized=0;

static int fbuf_page_order(unsigned long fbuf_pgs){
    int count = 0;

    if ( !fbuf_pgs ) 
            BUG();
    while ( count < 30 ){
       if ( (1<<count) >= fbuf_pgs)
            return count;
        count+=2;
    }
    BUG();
    return -EINVAL;
}

static __init void init_fbufs(void){
    struct page *pg; 
    struct page *page; 
    unsigned long lo0, lo1, hi;
    unsigned long fbuf_pgs;
    int fbuf_pg_order;
    int fbuf_chunk;
    int i;

    for ( fbuf_chunk = 0, i = 0; i < FBUF_MAX_TYPES; i++){
        fbuf_chunk += fbuf_ent_sz[i].count * fbuf_ent_sz[i].sz; 
    }

    fbuf_pgs = (fbuf_chunk/ PAGE_SIZE) + ((fbuf_chunk% PAGE_SIZE)? 1:0);
    fbuf_pg_order = fbuf_page_order(fbuf_pgs);
    fbuf_pgs = (1<<fbuf_pg_order);
    pg = alloc_pages(GFP_KERNEL,fbuf_pg_order);
	if (!pg)
            BUG();
    memset(page_address(pg),0,fbuf_pgs*PAGE_SIZE);
    
    /* Tell the system that we are reserving the address range*/
    if (remap_pages(FBUF_MAGIC_VMSTART, 
                        __pa(page_address(pg)), fbuf_pgs*PAGE_SIZE,
                        _PAGE_GLOBAL|_PAGE_VALID|__WRITEABLE)) {
        BUG();
    }

    /*Read Only segment*/
    if (remap_pages(FBUF_MAGIC_VMSTART + fbuf_pgs*PAGE_SIZE,
                        0, PAGE_SIZE,  _PAGE_GLOBAL)) {
        BUG();
    }
    
    /* Make a wired TLB*/
    hi = FBUF_MAGIC_VMSTART;
    lo0 = ((__pa(page_address(pg))>>6)&0x3fffffc0)|0x17;
    lo1 = 0x1;
    add_wired_entry(lo0,lo1,hi,(fbuf_pgs-1)<<13);

    /* Lock the pages and avoid swapping*/
    for (page = pg, i = 0; i < fbuf_pgs; i++, page++){
        if (TryLockPage(page))
            BUG();
    }

    /*zone management stuff allocation & initialization*/
    fbuf_zones = kmalloc(FBUF_MAX_TYPES * sizeof(struct fbuf_zone), GFP_KERNEL );
    if (!fbuf_zones)
        BUG();

    /*Initialize fastbuf_hdr*/
    strcpy(fastbuf->magic_word, FBUF_MAGIC_WORD);
    fastbuf->cur_pid = 0;
    fastbuf->rt_pri = 0;
    fastbuf->cur_exec_entry = NULL;
    fastbuf->rseq=1;
    //memset(&fastbuf->t,0,sizeof(monotonic_t));
    //memset(&fastbuf->tv,0,sizeof(struct timeval));
    
    /*Initializing zones*/
    fbuf_zones[0].zone_start = FBUF_MAGIC_VMSTART + sizeof(struct fastbuf_hdr);
    fbuf_zones[0].zone_end = fbuf_zones[0].zone_start + 
                                fbuf_ent_sz[0].count* fbuf_ent_sz[0].sz;

    fbuf_zones[0].in_use = 0;
    memset(fbuf_zones[0].bitmap,0,FBUF_MAX_ZBMP_BYTES);

    for ( i = 1; i < FBUF_MAX_TYPES; i++){
        fbuf_zones[i].zone_start = fbuf_zones[i-1].zone_end;
        fbuf_zones[i].zone_end = fbuf_zones[i].zone_start + 
                                    fbuf_ent_sz[i].count* fbuf_ent_sz[i].sz;
        fbuf_zones[i].in_use = 0;
        memset(fbuf_zones[i].bitmap,0,FBUF_MAX_ZBMP_BYTES);
    }
    fastbuf_initialized = ENABLE_FASTBUFS;
}
module_init(init_fbufs);

static int find_fbuf_zone(void *ptr){
    int zone;
    if (!ptr)
        return -EINVAL;
    
    for ( zone = 0; zone < FBUF_MAX_TYPES; zone++){
        if ( (unsigned long)ptr >= fbuf_zones[zone].zone_start && 
                (unsigned long)ptr < fbuf_zones[zone].zone_end )
                return zone;
    }
    return -EINVAL;
}

static void do_fbuf_free(void *ptr){
    int c,i,zone;
    zone = find_fbuf_zone(ptr);

    if ( zone < 0 )
        return;

    switch(zone){
        case FBUF_THREAD_EXEC_ENTRY:
          current->exec_cntxt = NULL;
          break;
        case FBUF_CNTXT_ENTRY:
          break;
        default:
          return;
    }

    c = ((unsigned long) ptr - fbuf_zones[zone].zone_start)/fbuf_ent_sz[zone].sz;
    if ( c < 0 )
        return ;
    
    i = c/8;
    if ( i < 0 )
        return ;

    if (fbuf_zones[zone].bitmap[i] & (1<<(c%8))){
            fbuf_zones[zone].bitmap[i] &= ~(1<<(c%8));
            fbuf_zones[zone].in_use--;
            if ( fbuf_zones[zone].in_use < 0)
                    BUG();
            dprintk("Fbuf Free: in_use %d bitmap (%d:%d):%08x\n",
                    fbuf_zones[zone].in_use,i,c%8,
                    fbuf_zones[zone].bitmap[i]);
            return;
    }
}

static void *do_fbuf_alloc(fbuf_type_t fbuf_zone){
    void *ptr;
    struct thread_exec_entry *e;
    struct cntxt_entry *cntxt;
    int c,i;

    if ( fbuf_zone > FBUF_MAX_TYPES )
            return NULL;

    for ( c = 0; c < FBUF_MAX_ZBMP_BYTES; c++){
        for ( i = 0; i < 8 ; i++){
            if (!(fbuf_zones[fbuf_zone].bitmap[c] & (1<<i))){
                    ptr = (void *)(fbuf_zones[fbuf_zone].zone_start + 
                            (((c*8)+i) * fbuf_ent_sz[fbuf_zone].sz));
                    fbuf_zones[fbuf_zone].bitmap[c] |= (1<<i);
                    fbuf_zones[fbuf_zone].in_use++;
                    dprintk("(%d:%d)ptr %08x fbuf_zone %d in_use %d bitmap %08x "
                            "start %08x end %08x\n",
                            i,c,ptr,fbuf_zone, 
                            fbuf_zones[fbuf_zone].in_use,fbuf_zones[fbuf_zone].bitmap[c],
                            fbuf_zones[fbuf_zone].zone_start, fbuf_zones[fbuf_zone].zone_end);
                    if ( (unsigned long)ptr > fbuf_zones[fbuf_zone].zone_end)
                            BUG();
                    if ( fbuf_zones[fbuf_zone].in_use > FBUF_MAX_ENTS)
                            BUG();
                   goto found; 
            }
        }
    }

found:
    if ( c >= FBUF_MAX_ZBMP_BYTES )
            return NULL;

    switch ( fbuf_zone ){

        case FBUF_THREAD_EXEC_ENTRY:
          e = (struct thread_exec_entry *) ptr;
          if( current->exec_cntxt != NULL ){
              do_fbuf_free(ptr);
              return NULL;
          }
          memset(e, 0, THREAD_EXEC_ENTRY_SZ);
          e->pid = current->pid;
          current->exec_cntxt = e;
        break;
          
        case FBUF_CNTXT_ENTRY:
          cntxt = (struct cntxt_entry *)ptr;
          memset(cntxt, 0, CNTXT_ENTRY_SZ);
        break;

        default:
          ptr = NULL;
        break;
    }
    return ptr;
}

#define FBUF_MAJOR 242
int fbuf_open (struct inode *inode, struct file *filp){
    return 0;
}

int fbuf_ioctl (struct inode *inode, struct file *file, 
            unsigned int cmd, unsigned long arg) {
    int err;
    struct fbuf_req fbuf_req;
    

    if (!fastbuf_initialized){
        printk(KERN_ERR "FastBuffers are not initialized\n");
        return -EINVAL;
    }

    if(_IOC_TYPE(cmd) != FBUF_IOC_MAGIC) {
        dprintk(KERN_ERR "[fbuf_ioctl] cmd (0x%x) mismatch, "
                        "called with 0x%x while expecting 0x%x\n", 
                        cmd, _IOC_TYPE(cmd), FBUF_IOC_MAGIC);
        return -EINVAL;
    }

    switch(cmd){
        case FBUF_ALLOC:
            err = copy_from_user((void *)&fbuf_req,(void *)arg,sizeof(struct fbuf_req));
            if ( err )
                return -EFAULT;
            fbuf_req.ptr = do_fbuf_alloc(fbuf_req.fbuf_type);
            err = copy_to_user((void *)arg,(void *)&fbuf_req,sizeof(struct fbuf_req));
            if ( err )
                return -EFAULT;
        break;

        case FBUF_FREE:
            do_fbuf_free((void *)arg);
        break;

        default:
            printk(KERN_ERR "fbuf_ioctl: Invalid command\n");
            return -EINVAL;
    }
    return 0;
}

static struct file_operations fbuf_fops = {
    ioctl:   fbuf_ioctl,
    open:    fbuf_open,
};

void __init init_fbuf_if(void){
    int result;
    result = register_chrdev(FBUF_MAJOR, "fbuf", &fbuf_fops);
    if ( result < 0 )
            printk(KERN_ERR "fbuf: "
                            "failed to register char dev [major %d]\n",FBUF_MAJOR);
}

module_init(init_fbuf_if);

void *alloc_fbuf_thread_exec_entry(void){
    return do_fbuf_alloc(FBUF_THREAD_EXEC_ENTRY);
}

void free_fbuf_thread_exec_entry(void *ptr){
    do_fbuf_free(ptr);
}
