/*
 * Support for periodic interrupts (100 per second) and for getting
 * the current time from the RTC on Power Macintoshes.
 *
 * We use the decrementer register for our periodic interrupts.
 * On processors other than the 601, we calibrate the decrementer
 * frequency using the VIA timer 1.
 *
 * Paul Mackerras	August 1996.
 * Copyright (C) 1996 Paul Mackerras.
 *
 * Mostly not his code anymore; thanks for the names
 * Copyright (C) 1998 Teleworld, Inc. All Rights Reserved.
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/ppcops.h>

#ifdef CONFIG_TCD_EVAL
#define TCD_CLOCK	(33 * 2 * 1000000)	/* clock doubled 33MHz */
#else
#define TCD_CLOCK	(27 * 2 * 1000000)	/* clock doubled 27MHz */
#endif
#define USECS_PER_JIFFY (1000000/HZ)

unsigned decrementer_count	= TCD_CLOCK / HZ;
unsigned counts_per_usec	= TCD_CLOCK / 1000000;
extern volatile unsigned long lost_ticks;

/*
 * This version of gettimeofday has microsecond resolution.
 */
void
do_gettimeofday(struct timeval *tv)
{
	int i;
	unsigned long flags;

	save_flags(flags);
	cli();
	*tv = xtime;
	tv->tv_usec += (decrementer_count - mfspr(SPRN_PIT)) / counts_per_usec;
	i = lost_ticks;
	while (i--)
		tv->tv_usec += USECS_PER_JIFFY;
	restore_flags(flags);

	while (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
}

void
do_settimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned int usec;

	save_flags(flags);
	cli();
	usec = (decrementer_count - mfspr(SPRN_PIT)) / counts_per_usec;
	xtime.tv_sec = tv->tv_sec;
	xtime.tv_usec = tv->tv_usec - usec;
	restore_flags(flags);
}

/*
 * timer_interrupt - gets called when the decrementer overflows,
 * with interrupts disabled.
 * We set it up to overflow again in 1/HZ seconds.
 */
void timer_interrupt(struct pt_regs * regs)
{
	mtspr(SPRN_TSR, 0x8000000);	/* acknowledge interrupt */
	do_timer(regs);
}

void __init time_init(void)
{
	int oiocr;

	xtime.tv_sec = 0; /* XXX */
	xtime.tv_usec = 0;

 	/* insure running clock-doubled */
 	oiocr = mfdcr(SPRN_IOCR);
 	oiocr |= 0x00004000;
 	mtdcr(SPRN_IOCR, oiocr);

	mtspr(SPRN_TCR, 0x04400000);	/* enable auto reload */
	mtspr(SPRN_TSR, 0xcc00000);	/* clear any lingering state */
 	mtspr(SPRN_PIT, decrementer_count);
}
