#ifndef	_ASM_KDB_H
#define _ASM_KDB_H

#include <asm/addrspace.h>
/*
 * Minimalist Kernel Debugger
 *
 * Copyright 2002 Intransa, Inc.
 *
 * asm-mips/kdb.h first written by Paul Lai
 *   heavily leveraged from asm-i386/kdb.h
 *
 */

#define KDBSW_BP	12	/* SW breakpoint */
#define KDBENTRY_BP	13	/* Initial KDB entry break point */
	/*
	 * KDB_ENTER() is a macro which causes entry into the kernel
	 * debugger from any point in the kernel code stream.  If it 
	 * is intended to be used from interrupt level, it must  use
	 * a non-maskable entry method.
	 */
#define KDB_ENTER()	asm("\tbreak 13\n")	/* must match KDBENTRY_BP */

	/*
	 * Define the exception frame for this architeture
	 */
struct pt_regs;
typedef struct pt_regs	*kdb_eframe_t;

	/*
	 * Needed for exported symbols.
	 */
typedef unsigned long kdb_machreg_t;

#define kdb_machreg_fmt		"0x%lx"
#define kdb_machreg_fmt0	"0x%08lx"
#define kdb_bfd_vma_fmt		"0x%lx"
#define kdb_bfd_vma_fmt0	"0x%08lx"
#define kdb_elfw_addr_fmt	"0x%x"
#define kdb_elfw_addr_fmt0	"0x%08x"

	/*
	 * Per cpu arch specific kdb state.  Must be in range 0xff000000.
	 */
#define KDB_STATE_A_IF		0x01000000	/* Saved IF flag */

	/*
	 * Functions to safely read and write kernel areas.  The {to,from}_xxx
	 * addresses are not necessarily valid, these functions must check for
	 * validity.  If the arch already supports get and put routines with
	 * suitable validation and/or recovery on invalid addresses then use
	 * those routines, otherwise check it yourself.
	 */

	/*
	 * asm-i386 uaccess.h supplies __copy_to_user which relies on MMU to
	 * trap invalid addresses in the _xxx fields.  Verify the other address
	 * of the pair is valid by accessing the first and last byte ourselves,
	 * then any access violations should only be caused by the _xxx
	 * addresses,
	 */

#include <asm/uaccess.h>

static inline int
kdba_putarea_size(unsigned long to_xxx, void *from, size_t size)
{
	mm_segment_t oldfs = get_fs();
	int r;
#if 0
	char c;
	c = *((volatile char *)from);
	c = *((volatile char *)from + size - 1);
	return __copy_to_user((void *)to_xxx, from, size);
#endif
	if (to_xxx < KSEG0)
		return -1;
	// memcpy((void*) to_xxx, from, size);
	r = __copy_to_user((void *)to_xxx, from, size);
	set_fs(oldfs);
	return (r);
}

static inline int
kdba_getarea_size(void *to, unsigned long from_xxx, size_t size)
{
	mm_segment_t oldfs = get_fs();
	int r;
#if 0
	*((volatile char *)to) = '\0';
	*((volatile char *)to + size - 1) = '\0';
	return __copy_to_user(to, (void *)from_xxx, size);
#endif
	if (from_xxx < KSEG0)
		return -1;
	// memcpy(to, (void*) from_xxx, size);
	r =  __copy_to_user(to, (void *)from_xxx, size);
	set_fs(oldfs);
	return (r);
}

static inline int
kdba_verify_rw(unsigned long addr, size_t size)
{
	unsigned char data[size];
	return(kdba_getarea_size(data, addr, size) || kdba_putarea_size(addr, data, size));
}

#endif	/* !_ASM_KDB_H */
